__author__ = 'jonpsull'

from es_logger.interface import ConsoleLogEventRegex
from es_logger.interface import EventGenerator
import json
import logging
import os
import re
from stevedore import driver, ExtensionManager

LOGGER = logging.getLogger(__name__)


class ConsoleLogEvent(EventGenerator):
    """
    Load regex's and scan the console log for matches, generating events
    """
    def get_fields(self):
        return super(ConsoleLogEvent, self).get_fields() + ['NODE_NAME', 'NODE_LABELS']

    def generate_events(self, esl):
        """
        Parse the console log and return a list to send as events

        :param console_log: A dictionary with string keys and simple types as
        :type console_log: str
        :returns: list(dict(str:?))
        """
        LOGGER.debug("Starting: {}".format(type(self).__name__))
        output_list = []
        regex_list = []

        """
        Load all console log plugins
        """
        mgr = ExtensionManager(namespace='es_logger.plugins.event_generator.console_log_events',
                               invoke_on_load=False)
        for plugin in mgr.names():  # All known console_log_events plugins
            drv = driver.DriverManager(
                namespace='es_logger.plugins.event_generator.console_log_events',
                invoke_on_load=False,
                name=plugin)
            regex_list = drv.driver.get_regex(regex_list)
            LOGGER.debug('Loaded plugins from {}, regex_list now with {} elements'.format(
                         plugin, len(regex_list)))

        """
        Look for a json file with additional regex's
        """
        regex_filename = 'console_log_event_regex.json'
        if os.path.isfile(regex_filename):
            LOGGER.debug("Loading console_log_regex's from {}".format(regex_filename))
            with open(regex_filename, 'r') as regex_file:
                regex_data = json.loads(regex_file)
                for regex in regex_data:
                    regex_list.append((regex['name'], re.compile(regex['pattern'], re.MULTILINE)))
        else:
            LOGGER.debug("{} not found, not loading additional console_log_event_regex's".format(
                         regex_filename))

        """
        Processing
        """
        for regex_name, run_regex in regex_list:
            LOGGER.debug("Starting regex: {}".format(regex_name))
            for regex_match in run_regex.finditer(esl.console_log):
                add_event = {
                            'name': regex_name,
                            'match': regex_match.group(0),
                            'named_matches': regex_match.groupdict()
                        }
                output_list.append(add_event)
            LOGGER.debug("Finished regex: {}".format(regex_name))

        LOGGER.debug("Finished: {}".format(type(self).__name__))
        # Return the data
        return output_list


class EsLoggerConsoleLogRegex(ConsoleLogEventRegex):
    """
    """
    def get_regex(regex_list):
        """
        Pipeline job information
        """
        # Collect agent information for this job
        regex_list.append(("jenkins agent", re.compile(
            r"Running on (?P<jenkins_agent>.*?) in (?P<workspace>/.*?)\s*\n+",
            re.MULTILINE)))

        # Collect stopwatch results
        regex_list.append(("stopwatch", re.compile(
            r"stopwatch - name:(?P<name>.*?), elapsedTime:(?P<elapsed_time>.*?), "
            r"startTime:(?P<start_time>.*?), endTime:(?P<end_time>.*?), "
            r"status:(?P<status>.*?)\s*\n+",
            re.MULTILINE)))

        """
        Docker errors
        """
        # Generic error - should pick all docker errors, would be great to auto-classify somehow
        regex_list.append(("docker error", re.compile(
            r"Error response from daemon:(?P<docker_error>.*?)\n+",
            re.MULTILINE)))

        # Registry issue
        regex_list.append(("docker blob error", re.compile(
            r"\n(?P<preceding>.*?)\n^(?P<blob>.*unknown blob.*)$",
            re.MULTILINE)))

        # Registry issue
        regex_list.append(("docker timeout", re.compile(
            r"Error response from daemon:(?P<err>.*?)net/http: request canceled " +
            r"\(Client\.Timeout exceeded while awaiting headers\)",
            re.MULTILINE)))

        # Registry issue
        regex_list.append(("docker missing signature", re.compile(
            r"Error response from daemon: missing signature key",
            re.MULTILINE)))

        # Registry issue
        regex_list.append(("docker image not found", re.compile(
            r"Pulling repository (?P<docker_repository>\S+)\n" +
            r"Error: image (?P<docker_image>\S+) not found",
            re.MULTILINE)))

        # Might be replaced by the multiline match, above
        # Registry issue
        regex_list.append(("docker image not found", re.compile(
            r"Error: image (?P<image>\S+) not found",
            re.MULTILINE)))

        # Registry issue
        regex_list.append(("docker pull error", re.compile(
            r"Error pulling image \((?P<docker_tag>\S+)\) from (?P<docker_registry>\S+), " +
            r"endpoint: (?P<docker_endpoint>\S+), (?P<error>.*?)\n",
            re.MULTILINE)))

        # Registry issue
        regex_list.append(("docker pull error", re.compile(
            r"Please login prior to pull:(?:\s*?\n*)*?" +
            r"Error: Cannot perform an interactive login from a non TTY device",
            re.MULTILINE)))

        # Registry issue
        regex_list.append(("http_status", re.compile(
            r"received unexpected HTTP status:(?P<http_status>.*?)\n+",
            re.MULTILINE)))

        # Client/agent issue
        regex_list.append(("thin pool full", re.compile(
            r"devmapper: Thin Pool has [0-9]+ free data blocks which is less than minimum " +
            r"required [0-9]+ free data blocks. Create more free space in thin pool or use " +
            r"dm.min_free_space option to change behavior",
            re.MULTILINE)))

        """
        Docker informational
        """
        regex_list.append(("reclaimed space", re.compile(
            r"Total reclaimed space.*?B",
            re.MULTILINE)))

        """
        Unclassified error
        """
        regex_list.append(("stdout temporarily unavailable", re.compile(
            r"/dev/stdout: resource temporarily unavailable",
            re.MULTILINE)))

        """
        Generic Jenkins errors
        """
        regex_list.append(("java exception", re.compile(
            r"^(?P<exception>java[.].*Exception):\s*(?P<exception_text>.*?)$",
            re.MULTILINE)))

        regex_list.append(("jenkins error", re.compile(
            r"^ERROR:\s*(?P<error_text>.*?)\n.*?\n+",
            re.MULTILINE)))

        regex_list.append(("jenkins fatal error", re.compile(
            r"^FATAL:\s*(?P<fatal_text>.*?)\n.*?\n+",
            re.MULTILINE)))

        regex_list.append(("remote file operation failed", re.compile(
            r"^remote file operation failed:\s*(?P<remote_fail_text>.*?)\n.*?\n+",
            re.MULTILINE)))

        return regex_list



# Copyright (c) 2018 Cisco Systems, Inc.
# All rights reserved.

__author__ = 'jonpsull'

import es_logger
import nose
import unittest.mock


class TestCommitEvent(object):
    def test_commit_event(self):
        eg = es_logger.plugins.commit.CommitEvent()
        fields = eg.get_fields()
        nose.tools.ok_(fields == es_logger.interface.EventGenerator.DEFAULT_FIELDS)
        esl = unittest.mock.MagicMock()
        esl.build_info = {
                "changeSet": {
                    "_class": "hudson.plugins.git.GitChangeSetList",
                    "items": [{"date": "2018-03-07 09:22:36 +0000"}],
                    "kind": "git"},
                "changeSets": [{
                    "items": [{"date": "2018-03-14T12:09:16+0000"}],
                    "kind": "git"}]}
        events = eg.generate_events(esl)
        nose.tools.ok_(len(events) == 2,
                       "Wrong number of events returned: {}".format(events))
        nose.tools.ok_(events[0]['changeSet']['date'] == "2018-03-14T12:09:16+0000",
                       "Bad event[0] returned: {}".format(events))
        nose.tools.ok_(events[1]['changeSet']['date'] == "2018-03-07T09:22:36+0000",
                       "Bad event[1] returned: {}".format(events))



# Copyright (c) 2018 Cisco Systems, Inc.
# All rights reserved.

__author__ = 'jonpsull'

from ..interface import EventGenerator


class CommitEvent(EventGenerator):
    """
    """
    def __init__(self):
        super().__init__()

    def get_fields(self):
        return super().get_fields()

    def generate_events(self, esl):
        # Prep the event info if there are change sets
        add_events = []
        parse_events = []
        # Process the change sets and create events for each - pipeline job
        for change_set in esl.build_info.get('changeSets', []):
            parse_events += change_set['items']
        if 'changeSet' in esl.build_info:
            parse_events += esl.build_info['changeSet']['items']

        for change_item in parse_events:
            new_event = {'changeSet': change_item}
            new_event['build_data'] = esl.es_info.get('build_data')
            # If the date is in a bad format, modify it
            if 'date' in new_event['changeSet'].keys() and \
                    ' ' in new_event['changeSet']['date']:
                parts = new_event['changeSet']['date'].split()
                new_event['changeSet']['date'] = parts[0] + 'T' + parts[1] + parts[2]
            add_events.append(new_event)
        return add_events



# Copyright (c) 2018 Cisco Systems, Inc.
# All rights reserved.

__author__ = 'jonpsull'

import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
version_file = os.path.join(here, 'VERSION')

# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

if not os.path.isfile(version_file):
    BUILD_NUMBER = os.environ.get('TRAVIS_BUILD_NUMBER', '0')
    with open(version_file, encoding='utf-8', mode='w') as f:
        f.write(BUILD_NUMBER)
else:
    with open(version_file, encoding='utf-8') as f:
        BUILD_NUMBER = f.read().strip()

setup(
    name='es_logger',
    version='2.' + BUILD_NUMBER,

    description='Framework for Creating Logstash events from Jenkins Jobs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/CiscoDevNet/es-logger',

    author='JP Sullivan (j3p0uk)',
    author_email='jonpsull@cisco.com',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Testing',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],

    keywords='jenkins development elasticsearch logstash build',

    install_requires=[
        'python-jenkins>=1.0.0',
        'pyzmq',
        'requests',
        'stevedore',
    ],

    platforms=['Any'],

    scripts=[],

    provides=['es_logger',
              'zmq_es_logger'],

    packages=find_packages(exclude=['docs', 'test']),
    include_package_data=True,

    entry_points={
        'console_scripts': [
            'es-logger=es_logger.cli:main',
            'zmq-es-logger=es_logger.zmq_client:main',
        ],
        'es_logger.plugins.console_log_processor': [
        ],
        'es_logger.plugins.gather_build_data': [
        ],
        'es_logger.plugins.event_generator': [
            'ansible_fatal = es_logger.plugins.ansible:AnsibleFatalGenerator',
            'ansible_recap_v2 = es_logger.plugins.ansible:AnsibleRecapEvent',
            'commit = es_logger.plugins.commit:CommitEvent',
            'console_log_events = es_logger.plugins.console_log_events:ConsoleLogEvent',
            'junit = es_logger.plugins.junit:JUnitEvent',
            'stages = es_logger.plugins.stages:StageEvent',
        ],
        'es_logger.plugins.event_generator.console_log_events': [
            'es_logger = es_logger.plugins.console_log_events:EsLoggerConsoleLogRegex',
            'ansible = es_logger.plugins.ansible:AnsibleConsoleLogRegex',
        ],
        'es_logger.plugins.event_target': [
            'logstash = es_logger.plugins.target:LogstashTarget',
        ],
    },

    data_files=[("", ["LICENSE", "VERSION"])],

    project_urls={
        'Bug Reports': 'https://github.com/CiscoDevNet/es-logger/issues',
        'Source': 'https://github.com/CiscoDevNet/es-logger/',
    },

    zip_safe=False,
)



# Copyright (c) 2018 Cisco Systems, Inc.
# All rights reserved.

__author__ = 'mvillene'

import es_logger
import nose
import unittest.mock


class TestStageEvent(object):
    def test_stage_fields(self):
        eg = es_logger.plugins.stages.StageEvent()
        fields = eg.get_fields()
        nose.tools.ok_(fields == es_logger.interface.EventGenerator.DEFAULT_FIELDS)

    def test_no_stages(self):
        eg = es_logger.plugins.stages.StageEvent()
        esl = unittest.mock.MagicMock()
        esl.get_stages.return_value = None
        events = eg.generate_events(esl)
        nose.tools.ok_(len(events) == 0)

    def test_with_stage_report(self):
        eg = es_logger.plugins.stages.StageEvent()
        esl = unittest.mock.MagicMock()
        esl.get_stages.return_value = {
            "_links": {"self": {"href": "/job/test/job/HelloWorld/78/wfapi/describe"}},
            "id": "78", "name": "#78", "status": "SUCCESS", "startTimeMillis": 1549926641427,
            "endTimeMillis": 1549926650050, "durationMillis": 8623, "queueDurationMillis": 1,
            "pauseDurationMillis": 0,
            "stages": [
                {"_links": {"self": {
                    "href": "/job/test/job/HelloWorld/78/execution/node/6/wfapi/describe"}},
                 "id": "6", "name": "Stage 1", "execNode": "", "status": "SUCCESS",
                 "startTimeMillis": 1549926642588, "durationMillis": 18, "pauseDurationMillis": 0},
                {"_links": {"self": {
                    "href": "/job/test/job/HelloWorld/78/execution/node/11/wfapi/describe"}},
                 "id": "11", "name": "Stage 2", "execNode": "", "status": "SUCCESS",
                 "startTimeMillis": 1549926642613, "durationMillis": 17, "pauseDurationMillis": 0},
                {"_links": {"self": {
                    "href": "/job/test/job/HelloWorld/78/execution/node/16/wfapi/describe"}},
                 "id": "16", "name": "Declarative: Post Actions", "execNode": "",
                 "status": "SUCCESS", "startTimeMillis": 1549926642637, "durationMillis": 7398,
                 "pauseDurationMillis": 0}
            ]
        }
        events = eg.generate_events(esl)
        nose.tools.ok_(len(events) == 3,
                       "Wrong number of events returned ({}): {}".format(len(events), events))
        results = [
            {"status": "SUCCESS", "id": "6", "pauseDurationMillis": 0, "durationMillis": 18,
             "_links": {"self": {
                "href": "/job/test/job/HelloWorld/78/execution/node/6/wfapi/describe"}},
             "startTimeMillis": 1549926642588, "execNode": "", "name": "Stage 1"},
            {"status": "SUCCESS", "id": "11", "pauseDurationMillis": 0, "durationMillis": 17,
             "_links": {"self": {
                "href": "/job/test/job/HelloWorld/78/execution/node/11/wfapi/describe"}},
             "startTimeMillis": 1549926642613, "execNode": "", "name": "Stage 2"},
            {"status": "SUCCESS", "id": "16", "pauseDurationMillis": 0, "durationMillis": 7398,
             "_links": {"self": {
                "href": "/job/test/job/HelloWorld/78/execution/node/16/wfapi/describe"}},
             "startTimeMillis": 1549926642637, "execNode": "", "name": "Declarative: Post Actions"}
        ]

        for idx, event in enumerate(events):
            nose.tools.ok_(event == results[idx], "Bad event[{}] returned: {}".format(idx, events))



# Copyright (c) 2018 Cisco Systems, Inc.
# All rights reserved.

__author__ = 'jonpsull'

from es_logger.plugins.target import LogstashPostError, LogstashTarget
import nose
import os
import requests
import unittest.mock


class TestLogstashTarget(object):

    def setup(self):
        self.lt = LogstashTarget()

    @unittest.mock.patch.dict('os.environ', {'LOGSTASH_SERVER': "https://example.com",
                                             'LS_USER': 'user',
                                             'LS_PASSWORD': 'pass'})
    def test_LogstashTarget(self):
        lt = LogstashTarget()
        nose.tools.ok_(lt.get_logstash_server() == os.getenv('LOGSTASH_SERVER'))
        nose.tools.ok_(lt.get_ls_user() == os.getenv('LS_USER'))
        nose.tools.ok_(lt.get_ls_password() == os.getenv('LS_PASSWORD'))

    @unittest.mock.patch('requests.Session')
    def test_get_session(self, mock_session):
        self.lt.get_session()
        print(mock_session.mock_calls)
        nose.tools.ok_(self.lt.ls_session is not None,
                       "Session is None: {}".format(self.lt.ls_session))

    @unittest.mock.patch('requests.Session')
    def test_send_event_good(self, mock_session):
        mock_session().post().ok = True
        res = self.lt.send_event({"event": "event"})
        nose.tools.ok_(res == 0,
                       "res not 0: {}".format(res))

    @unittest.mock.patch('requests.Session')
    def test_send_event_bad(self, mock_session):
        mock_session().post().ok = False
        res = self.lt.send_event({"event": "event"})
        nose.tools.ok_(res == 1,
                       "res not 1: {}".format(res))

    @unittest.mock.patch('requests.Session')
    def test_send_event_bad_session(self, mock_session):
        mock_session().post.side_effect = requests.exceptions.ReadTimeout
        self.lt.timeout_sleep = 0
        nose.tools.assert_raises(LogstashPostError, self.lt.send_event, {"event": "event"})
        # We should see this created and called 5 times attempting to get post finished
        calls = [unittest.mock.call(),
                 unittest.mock.call(),
                 unittest.mock.call().post(None, json={'event': 'event'}),
                 unittest.mock.call(),
                 unittest.mock.call().post(None, json={'event': 'event'}),
                 unittest.mock.call(),
                 unittest.mock.call().post(None, json={'event': 'event'}),
                 unittest.mock.call(),
                 unittest.mock.call().post(None, json={'event': 'event'}),
                 unittest.mock.call(),
                 unittest.mock.call().post(None, json={'event': 'event'})]
        print(mock_session.mock_calls)
        nose.tools.ok_(mock_session.mock_calls == calls)

    @unittest.mock.patch('requests.Session')
    def test_send_event_bad_post(self, mock_session):
        mock_session().post.side_effect = Exception
        self.lt.timeout_sleep = 0
        nose.tools.assert_raises(LogstashPostError, self.lt.send_event, {"event": "event"})

    def test_validate(self):
        ret = self.lt.validate()
        nose.tools.ok_(ret, "Validate must return True")
